#include <stdio.h>
#include <errno.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include "config.h"
#include "imapfilter.h"
#include "file.h"
#include "log.h"


account_data account;
filter_entry **filters;
int fcnt = 0;			/* Counter of how many filters were
				   found in configuration file. */


/*
 * Finds path to configuration file, opens it and calls parse_config().
 */
int read_config(char *cfg)
{
    int r;
    FILE *fp;
    char *home = NULL;

    if (!cfg) {
	cfg = (char *) malloc(PATH_MAX * sizeof(char));

	home = getenv("HOME");

	snprintf(cfg, PATH_MAX, "%s/%s", home, ".imapfilterrc");
    }
#ifdef DEBUG
    printf("debug: configuration file: %s\n", cfg);
#endif

    fp = fopen(cfg, "r");

    if (!fp) {
	fprintf(stderr, "imapfilter: Could not open config file %s; %s\n",
		cfg, strerror(errno));
	return FAILURE;
    }

    account.port = 143;		/* Default IMAP protocol port to connect. */

    r = parse_config(fp);

    fclose(fp);

    return r;
}

/*
 * Parses configuration file.
 */
int parse_config(FILE * fp)
{
    int i;
    unsigned int row = 0;
    char line[LINE_MAX];
    regex_t comreg[4];
    regmatch_t match[3];
    const char *reg[] = {
	"^[[:blank:]]*(SERVER|PORT|USERNAME|PASSWORD|LOGFILE)[[:blank:]]*=[[:blank:]]*([[:graph:]]*)[[:blank:]]*",
	"^[[:blank:]]*DENY[[:blank:]]*=[[:blank:]]*(FROM|CC|BCC|SUBJECT|TO):? ([[:print:]]*)",
	"^[[:blank:]]*DENY[[:blank:]]*=[[:blank:]]*([[:graph:]]+):? ([[:print:]]*)",
	"^[[:blank:]]*#{0,1}.*"
    };

    for (i = 0; i < 4; i++)
	regcomp(&comreg[i], reg[i], REG_EXTENDED);

    while (fgets(line, LINE_MAX - 1, fp)) {
	row++;
	if (!regexec(&comreg[0], line, 3, match, 0)) {
	    set_account(line, match);
	    continue;
	} else if (!regexec(&comreg[1], line, 3, match, 0)) {
	    standard_filter(line, match);
	    continue;
	} else if (!regexec(&comreg[2], line, 3, match, 0)) {
	    explicit_filter(line, match);
	    continue;
	} else if (!regexec(&comreg[3], line, 1, match, 0))
	    continue;
	else {
	    fprintf(stderr,
		    "imapfilter: parse error in config file at row %d\n",
		    row);
	    return FAILURE;
	}
    }
#ifdef DEBUG
    printf("debug: account setting SERVER: '%s'\n"
	   "debug: account setting PORT: %d\n"
	   "debug: account setting USERNAME: '%s'\n"
	   "debug: account setting PASSWORD: '%s'\n"
	   "debug: account setting LOGFILE: '%s'\n",
	   account.server, account.port, account.username,
	   account.password, account.logfile);
#endif
    return SUCCESS;
}


/*
 * An account setting was found from parse_config(), changes the
 * apropriate variable.
 */
void set_account(char *line, regmatch_t * match)
{
    int s;
    char port[6];

    if (!strncmp(line + match[1].rm_so, "SERVER", 6)) {
	s = min((match[2].rm_eo - match[2].rm_so), SERVER_MAX - 1);
	strncpy(account.server, line + match[2].rm_so, s);
	account.server[s] = 0;
    } else if (!strncmp(line + match[1].rm_so, "PORT", 4)) {
	strncpy(port, line + match[2].rm_so,
		min((match[2].rm_eo - match[2].rm_so), sizeof(port)));
	account.port = strtoul(port, NULL, 0);
    } else if (!strncmp(line + match[1].rm_so, "USERNAME", 8)) {
	s = min((match[2].rm_eo - match[2].rm_so), USERNAME_MAX - 1);
	strncpy(account.username, line + match[2].rm_so, s);
	account.username[s] = 0;
    } else if (!strncmp(line + match[1].rm_so, "PASSWORD", 8)) {
	s = min((match[2].rm_eo - match[2].rm_so), PASSWORD_MAX - 1);
	strncpy(account.password, line + match[2].rm_so, s);
	account.password[s] = 0;
    } else if (!strncmp(line + match[1].rm_so, "LOGFILE", 7)) {
	s = min((match[2].rm_eo - match[2].rm_so), PATH_MAX - 1);
	strncpy(account.logfile, line + match[2].rm_so, s);
	account.logfile[s] = 0;
    }
}


/*
 * A filter entry was found from parse_config() proccesses it and saves it.
 */
void set_filters(char *line, regmatch_t * match, int expl)
{
    int s;

    /* Every FILTER_ENTRIES_MAX, space for more filters is realloc() -ed. */
    if (!fcnt)
	filters =
	    (filter_entry **) malloc(FILTER_ENTRIES_MAX *
				    sizeof(filter_entry *));
    else if (!(fcnt % FILTER_ENTRIES_MAX))
	filters = (filter_entry **) realloc(filters,
					   (((fcnt / FILTER_ENTRIES_MAX) +
					     1) * FILTER_ENTRIES_MAX) *
					   sizeof(filter_entry *));
    
    filters[fcnt] = (filter_entry *) malloc(sizeof(filter_entry));

    filters[fcnt]->explicit = expl;
    
    s = min((match[1].rm_eo - match[1].rm_so), FIELD_NAME_MAX - 1);
    strncpy(filters[fcnt]->name, line + match[1].rm_so, s);
    filters[fcnt]->name[s] = 0;

    s = min((match[2].rm_eo - match[2].rm_so), FIELD_BODY_MAX - 1);
    strncpy(filters[fcnt]->body, line + match[2].rm_so, s);
    filters[fcnt]->body[s] = 0;

#ifdef DEBUG
    printf("debug: filter entry '%s': '%s'\n", filters[fcnt]->name,
	   filters[fcnt]->body);
#endif

    fcnt++;
}
