#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "config.h"
#include "imapfilter.h"
#include "imap.h"
#include "connect.h"
#include "log.h"


static int tag = 0xA000;	/* Every IMAP command is prefixed with a
				   unique [:alnum:] string. */
extern int sock;
extern int fcnt;
extern filter_entry **filters;
extern account_data account;
extern int options;


/*
 * Sends to server data; a command.
 */
int send_command(char *command)
{

#ifdef DEBUG
    printf("debug: sending command: %s", command);
#endif

    if (write(sock, command, strlen(command)) == -1) {
	error("imapfilter: error while sending command; %s",
	      strerror(errno));
	return FAILURE;
    }

    return SUCCESS;
}


/*
 * Calls send_command() and get_response().
 */
int send_command_get_response(char *command)
{
    char buf[RESPONSE_BUFFER_MAX];

    if (!send_command(command)) {
	get_response(buf);
	return SUCCESS;
    } else
	return FAILURE;
}


#ifdef DEBUG
/*
 * IMAP NOOP: does nothing always succeeds.
 */
int imap_noop(void)
{
    char command[SMALL_COMMAND_MAX];

    verbose("Client request:  NOOP\n");

    snprintf(command, SMALL_COMMAND_MAX, "%X NOOP\r\n", tag++);

    return send_command_get_response(command);

}
#endif


/*
 * IMAP LOGOUT: informs server that client is done.
 */
int imap_logout(void)
{
    char command[SMALL_COMMAND_MAX];

    verbose("Client request:  LOGOUT\n");

    snprintf(command, SMALL_COMMAND_MAX, "%X LOGOUT\r\n", tag++);

    return send_command_get_response(command);
}


/*
 * IMAP LOGIN: identifies client to server.
 */
int imap_login(void)
{
    char command[MEDIUM_COMMAND_MAX];

    verbose("Client request:  LOGIN\n");

    snprintf(command, MEDIUM_COMMAND_MAX, "%X LOGIN \"%s\" \"%s\"\r\n",
	     tag++, account.username, account.password);

    return send_command_get_response(command);
}


/*
 * IMAP SELECT: selects a mailbox in which messages can be accessed.
 */
int imap_select(void)
{
    char command[SMALL_COMMAND_MAX];

    verbose("Client request:  SELECT\n");

    snprintf(command, SMALL_COMMAND_MAX, "%X SELECT INBOX\r\n", tag++);

    return send_command_get_response(command);
}

/*
 * IMAP SEARCH: searches the mailbox for messages that match certain criteria.
 */
int imap_search(void)
{
    char command[BIG_COMMAND_MAX];

    verbose("Client request:  SEARCH\n");

    generate_search_filters(command);

    return send_command(command);
}


/*
 * IMAP FETCH: retrieves data associated with a message.
 */
int imap_fetch(unsigned int m, char *results)
{
    char command[MEDIUM_COMMAND_MAX];

    verbose("Client request:  FETCH\n");

    snprintf(command, MEDIUM_COMMAND_MAX,
	     "%X FETCH %d BODY[HEADER.FIELDS (\"DATE\" \"FROM\" \"SUBJECT\")]\r\n",
	     tag++, m);

    if (!send_command(command)) {
	fetch_response(results);
	return SUCCESS;
    } else
	return FAILURE;
}


/*
 * IMAP STORE: alters data associated with a message.
 */
int imap_store(unsigned int m)
{
    char command[MEDIUM_COMMAND_MAX];

    verbose("Client request:  STORE\n");

    snprintf(command, MEDIUM_COMMAND_MAX,
	     "%X STORE %d +FLAGS \\Deleted\r\n", tag++, m);

    return send_command_get_response(command);
}


/*
 * IMAP EXPUNGE: premanently removes any messages with the \Deleted flag set.
 */
int imap_expunge(void)
{
    char command[SMALL_COMMAND_MAX];

    verbose("Client request:  EXPUNGE\n");

    snprintf(command, SMALL_COMMAND_MAX, "%X EXPUNGE\r\n", tag++);

    return send_command_get_response(command);
}


/*
 * Prepare according to the filters defined by the user the IMAP SEARCH
 * command that will be sent.
 */
void generate_search_filters(char *command)
{
    int f = 0;
    int len;

    snprintf(command, BIG_COMMAND_MAX, "%X SEARCH ", tag++);

    while (f < (fcnt - 1)) {
	len = strlen(command);
	snprintf(command + len, BIG_COMMAND_MAX - len, "OR %s%s \"%s\" ",
		 (filters[f]->explicit ? "HEADER " : ""), filters[f]->name,
		 filters[f]->body);
	f++;
    }

    len = strlen(command);
    snprintf(command + len, BIG_COMMAND_MAX - len, "%s%s \"%s\"\r\n",
	     (filters[f]->explicit ? "HEADER " : ""), filters[f]->name,
	     filters[f]->body);
}


/*
 * Convert the string with the UID's of the messages to be deleted to
 * integers, send the appropriate command and maybe print some headers
 * of the "to be deleted" message.
 */
void delete_messages(char *unwanted)
{
    unsigned long int messg;
    char headers[HEADERS_MAX];

    do {
	messg = strtoul(unwanted, &unwanted, 0);

	if (options & OPTION_SHOW_HEADERS) {
	    imap_fetch(messg, headers);
	    
	    if (!(options & OPTION_DETAILS_QUITE))
		printf("Deleting message:\n%s", headers);
	    	    
	    if (!(options & OPTION_TEST_MODE))
		log_info("%s", headers);
	}
	
	if (!(options & OPTION_TEST_MODE))
	    imap_store(messg);
	
    } while (*unwanted);
}
